/*
 * Copyright (c) 2020 Roc Streaming authors
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

//! @file roc_pipeline/converter_source.h
//! @brief Converter source pipeline.

#ifndef ROC_PIPELINE_CONVERTER_SOURCE_H_
#define ROC_PIPELINE_CONVERTER_SOURCE_H_

#include "roc_audio/channel_mapper_reader.h"
#include "roc_audio/iresampler.h"
#include "roc_audio/poison_reader.h"
#include "roc_audio/profiling_reader.h"
#include "roc_audio/resampler_map.h"
#include "roc_audio/resampler_profile.h"
#include "roc_audio/resampler_reader.h"
#include "roc_core/buffer_factory.h"
#include "roc_core/optional.h"
#include "roc_core/scoped_ptr.h"
#include "roc_pipeline/config.h"
#include "roc_sndio/isource.h"

namespace roc {
namespace pipeline {

//! Converter source pipeline.
//! @remarks
//!  - input: frames
//!  - output: frames
class ConverterSource : public sndio::ISource, public core::NonCopyable<> {
public:
    //! Initialize.
    ConverterSource(const ConverterConfig& config,
                    sndio::ISource& input_source,
                    core::BufferFactory<audio::sample_t>& buffer_factory,
                    core::IAllocator& allocator);

    //! Check if the pipeline was successfully constructed.
    bool valid();

    //! Get sample specification of the source.
    virtual audio::SampleSpec sample_spec() const;

    //! Get latency of the source.
    virtual core::nanoseconds_t latency() const;

    //! Check if the sink has own clock.
    virtual bool has_clock() const;

    //! Get current source state.
    virtual State state() const;

    //! Pause reading.
    virtual void pause();

    //! Resume paused reading.
    virtual bool resume();

    //! Restart reading from the beginning.
    virtual bool restart();

    //! Adjust source clock to match consumer clock.
    virtual void reclock(packet::ntp_timestamp_t timestamp);

    //! Read frame.
    virtual bool read(audio::Frame&);

private:
    core::Optional<audio::ChannelMapperReader> channel_mapper_reader_;

    core::Optional<audio::PoisonReader> resampler_poisoner_;
    core::Optional<audio::ResamplerReader> resampler_reader_;
    core::ScopedPtr<audio::IResampler> resampler_;

    core::Optional<audio::PoisonReader> pipeline_poisoner_;
    core::Optional<audio::ProfilingReader> profiler_;

    sndio::ISource& input_source_;
    audio::IFrameReader* audio_reader_;

    ConverterConfig config_;
};

} // namespace pipeline
} // namespace roc

#endif // ROC_PIPELINE_CONVERTER_SOURCE_H_
